/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2003 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#ifndef __adevs_models_h_
#define __adevs_models_h_
#include "adevs_config.h"
#include "adevs_bag.h"
#include "adevs_set.h"
#include "adevs_exceptions.h"
#include "adevs_types.h"
#include <algorithm>

ADEVS_NAMESPACE_BEGIN

/**
Model I/O is represented by PortValue pairs.
*/
class PortValue
{
	public:
		/// The value of this PortValue pair
		ADEVS_IO_TYPE value;
		/// The port on which the value appears
		ADEVS_PORT_TYPE port;
		PortValue()
		{
		}
		PortValue(ADEVS_PORT_TYPE port, ADEVS_IO_TYPE value):
		value(value),
		port(port)
		{
		}
		PortValue(const PortValue& src):
		value(src.value),
		port(src.port)
		{
		}
		const PortValue& operator=(const PortValue& src)
		{
			value = src.value;
			port = src.port;
			return *this;
		}
		~PortValue()
		{
		}
};

/**
Predeclare network and atomic model so types can be used as the type of
parent in the basic devs model and for type ID functions.  
*/
class devsn;
class atomic;
class netExec;

/**
The devs class provides basic operations for all devs models.
Note that models are derived from the object class.  Therefore
(if using the default configuration) models can, themselves,
be passed as I/O between models (this can be useful when building
some types of dynamic structure models).
*/
class devs: public object
{
	public:
		/// Default constructor
		devs():
		object(),
		parent(NULL),
		key(0)
		{
		}
		/// Copy constructor
		devs(const devs& src):
		object(src),
		parent(NULL),
		key(src.key)
		{
		}
		/// Assignment operator
		void operator=(const devs& src)
		{
			object::operator=(src);
			parent = NULL;
			key = src.key;
		}
		/**
		The init function should place the model into its initial state.
		The time t0 is the initial simulation time.
		*/
		virtual void init(ADEVS_TIME_TYPE t0) = 0;
		/// For use by the simulator.  Returns a pointer to the models parent.
		devsn* getParent() { return parent; }
		const devsn* getParent() const { return parent; }
		/**
		The model parent should be set when the model becomes a component in
		a network model.  A value of NULL indicates that the model has no parent.
		*/
		void setParent(devsn* model) { parent = model; }
		/// Destructor
		virtual ~devs()
		{
		}
		/**
		Returns NULL if this is not a network model and returns a pointer to
		itself otherwise.  This method is used to avoid a relatively expensive
		dynamic cast, which resulted in significant performance penalties.
		*/
		virtual devsn* typeIsNetwork() { return NULL; }
		/// Returns NULL if this is not a atomic model; returns itself otherwise
		virtual atomic* typeIsAtomic() { return NULL; }
		/// Returns NULL if this is not a network executive; returns itself otherwise
		virtual netExec* typeIsNetExec() { return NULL; }
		/// Get the cell space index key assigned to this model
		unsigned long get_cellspace_key() const { return key; }
		/// Set the cell space index key for this model (used by the cellSpace class)
		void set_cellspace_key(unsigned long cell_key) { key = cell_key; }
	private:
		devsn* parent;
		unsigned long key;
};

/**
EventReceiver objects are used to route I/O within a coupled model
(see the netExec class).
*/
class EventReceiver
{
	public:
		/// The destination model
		devs* model;
		/// The destination port
		ADEVS_PORT_TYPE port;
		EventReceiver():
		model(NULL)
		{
		}
		EventReceiver(devs* model, ADEVS_PORT_TYPE port):
		model(model),
		port(port)
		{
		}
		EventReceiver(const EventReceiver& src):
		model(src.model),
		port(src.port)
		{
		}
		const EventReceiver& operator=(const EventReceiver& src)
		{
			model = src.model;
			port = src.port;
			return *this;
		}
		bool operator==(const EventReceiver& other) const
		{
			return (model == other.model && port == other.port);
		}
		bool operator!=(const EventReceiver& other) const
		{
			return !((*this) == other);
		}
		~EventReceiver(){}
};

typedef struct
{
	int operator()(const EventReceiver& r) const { return (int)(r.model); }
} 
EventReceiverHashFunc;

/**
The port_iterator can be used to iterator over the values on a specific port
in an input bag (i.e. adevs_bag<PortValue> types).
*/
class port_iterator
{
	public:
		port_iterator(const port_iterator& src):
		x(src.x),
		pos(src.pos)
		{
		}
		/// Create an iterator for bag x that will iterate over values on port
		port_iterator(const adevs_bag<PortValue>& x, ADEVS_PORT_TYPE port):
		x(&x),
		pos(0),
		port(port)
		{
			while (pos < x.getSize())
			{
				if (x.get(pos).port == port) return;
				else pos++;
			}
		}
		/// Advance the iterator to the next value on the port
		void operator++() 
		{ 
			if (pos < 0) pos = 0;
			else pos++;
			while (pos < x->getSize())
			{
				if (x->get(pos).port == port) return;
				else pos++;
			}
		}
		/// Advance the iterator to the next value on the port
		void operator++(int i)
		{
			operator++();
		}
		/// Return the iterator to the previous value on the port
		void operator--()
		{
			if (pos < 0) pos = 0;
			else pos--;
			while (pos >= 0)
			{
				if (x->get(pos).port == port) return;
				else pos--;
			}
		}
		/// Return the iterator to the previous value on the port
		void operator--(int i)
		{
			operator--();
		}
		/// Get the current value of the iterator (i.e. the value on the port)
		ADEVS_IO_TYPE operator*() const { return (x->get(pos)).value; }
		/// Returns true if at the end of the input on the port
		bool end_of_input() const { return pos >= x->getSize(); }
		/// Returns true if at the beginning of the input on the port
		bool start_of_input() const { return pos < 0; }
		~port_iterator()
		{
		}
	private:
		const adevs_bag<PortValue>* x;
		int pos;
		ADEVS_PORT_TYPE port;
		void operator=(const port_iterator& src){}
		port_iterator(){}
};

/**
Base type for all atomic DEVS models.
*/
class atomic: public devs
{
	public:
		atomic():
		devs(),
		tN(ADEVS_INFINITY),
		tL(ADEVS_ZERO),
		t(ADEVS_ZERO),
		x(NULL),
		y(NULL)
		{
		}
		atomic(const atomic& src):
		devs(src),
		tN(src.tN),
		tL(src.tL),
		t(src.t),
		x(NULL),
		y(NULL)
		{
		}
		const atomic& operator=(const atomic& src)
		{
			devs::operator=(src);
			tN = src.tN;
			tL = src.tL;
			t = src.t;
			x = y = NULL;
			return *this;
		}
		/// Calls init() method.  timeCurrent() == t0 will be true.
		void init(ADEVS_TIME_TYPE t0)
		{
			init();
		}
		/// Initialization function
		virtual void init() = 0;
		/// Internal transition function
		virtual void delta_int() = 0;
		/// External transition function
		virtual void delta_ext(ADEVS_TIME_TYPE e, const adevs_bag<PortValue>& x) = 0;
		/// Confluent transition function
		virtual void delta_conf(const adevs_bag<PortValue>& x) = 0;
		/// Output function
		virtual void output_func(adevs_bag<PortValue>& y) = 0;
		/**
		Garbage collection function.  The PortValue objects in g are
		no longer in use by the simulation engine and should be disposed of. 
`		Note that the elements in g are only those PortValue objects produced as
		output by this model.
		*/
		virtual void gc_output(adevs_bag<PortValue>& g) = 0;
		virtual ~atomic()
		{
		}
		/// Adds the PortValue pair (port,value) to the bag y.
		static void output(ADEVS_PORT_TYPE port, ADEVS_IO_TYPE value, 
		adevs_bag<PortValue>& y)
		{
			y.add(PortValue(port,value));
		}
		/// Get the time of the next internal event
		ADEVS_TIME_TYPE timeNext() const { return tN; }
		/// Get the time of the last state transition
		ADEVS_TIME_TYPE timeLast() const { return tL; }
		/// Get the current simulation time
		ADEVS_TIME_TYPE timeCurrent() const { return t; }
		/// Get the elapsed time since tL
		ADEVS_TIME_TYPE elapsed() const { return t-tL; }
		/// Get the time remaining until the next internal event
		ADEVS_TIME_TYPE sigma() const { return tN-t; }
		/// Get the value of the time advance function
		ADEVS_TIME_TYPE ta() const { return tN-tL; }
		/**
		Sets the time of next event (timeNext()) to timeCurrent()+dt.
		Throws an exception if dt < ADEVS_ZERO.  This is the preferred method
		for scheduling an event.
		*/
		void hold(ADEVS_TIME_TYPE dt) 
		{ 
			if (dt == ADEVS_INFINITY) tN = ADEVS_INFINITY;
			else if (dt >= ADEVS_ZERO) tN = t+dt; 
			else THROW_BASE_EXCEPTION("time advance < ADEVS_ZERO")
		}
		/// Set the time of next event to ADEVS_INFINITY
		void passivate() { tN = ADEVS_INFINITY; }
		/// For use by the simulator.  Sets the current time.
		void setTimeCurrent(ADEVS_TIME_TYPE t_current) { t = t_current; }
		/// For use by the simulator.  Sets the next event time.
		void setTimeNext(ADEVS_TIME_TYPE t_next) { tN = t_next; }
		/// For use by the simulator.  Sets the last event time.
		void setTimeLast(ADEVS_TIME_TYPE t_last) { tL = t_last; }
		atomic* typeIsAtomic() { return this; }
		/// For use by the simulator. Get the output bag for this model.
		adevs_bag<PortValue>* get_output_bag() { return y; }
		/// For use by the simulator. Get the input bag for this model.
		adevs_bag<PortValue>* get_input_bag() { return x; }
		/// For use by the simulator. Assign an input bag to this model.
		void set_input_bag(adevs_bag<PortValue>* xx) { x = xx; }
		/// For use by the simulator. Assign an output bag to this model.
		void set_output_bag(adevs_bag<PortValue>* yy) { y = yy; }
	private:
		ADEVS_TIME_TYPE tN, tL, t;
		adevs_bag<PortValue>* x;
		adevs_bag<PortValue>* y;
};

/**
Base type for all network executive models.
*/
class netExec: public atomic
{
	public:
		netExec(devsn* parent = NULL):
		atomic()
		{
			setParent(parent);
		}
		netExec(const netExec& src, devsn* parent):
		atomic(src)
		{
			setParent(parent);
		}
		/**
		Implementations of this method should fill out the
		set c with all components models (but not the parent or
		network executive itself).
		*/
		virtual void getComponents(adevs_set<devs*>& c) = 0;
		/**
		An implementation should fill the EventReceiver bag r
		with all destinations to receive pv.value.
		*/
		virtual void route(const PortValue& pv, devs* model, 
		adevs_bag<EventReceiver>& r) = 0;
		/**
		The models in the set removed are no long members of this
		network executives component set (as per getComponents())
		and are no longer in use by the simulation engine.
		*/
		virtual void gc_models(adevs_set<devs*>& removed) = 0;
		/**
		This method returns a hint to the simulation engine. If the
		return value is true, the simulation engine can skip the check
		for new or removed components.  Otherwise, the simulator will
		perform the new/removed component check.  By default, this method
		will always return true.
		*/
		virtual bool componentSetChanged() const
		{
			return true;
		}
		/// Destructor
		virtual ~netExec()
		{
		}
		netExec* typeIsNetExec() { return this; }
	protected:
		netExec(const netExec& src):
		atomic(src)
		{
		}
		void operator=(const netExec& src)
		{
			atomic::operator=(src);
		}
};

/**
Base class for DEVS network models.
*/
class devsn: public devs
{
	public:
		devsn():
		devs()
		{
		}
		devsn(const devsn& src):
		devs(src)
		{
		}
		const devsn& operator=(const devsn& src)
		{
			devs::operator=(src);
			return *this;
		}
		/**
		Any class derived from the devsn class should call
		the parent init() method when the derived init() method
		is called.  The init() method initializes all
		component models and sets their parent field.
		The init() method assumes that the network executive
		has already been created (i.e. getNetExec() != NULL).
		*/
		void init(ADEVS_TIME_TYPE t0)
		{
			// Initialize the network executive
			getNetExec()->setParent(this);
			getNetExec()->setTimeCurrent(t0);
			getNetExec()->setTimeLast(t0);
			getNetExec()->init(t0);
			// Initialize the component models
			adevs_set<devs*> models;
			getNetExec()->getComponents(models);
			for (int i = 0; i < models.getSize(); i++)
			{
				models[i]->setParent(this);
				atomic* amodel = models[i]->typeIsAtomic();
				if (amodel != NULL)
				{
					amodel->setTimeCurrent(t0);
					amodel->setTimeLast(t0);
				}
				models[i]->init(t0);
			}
		}
		/// Return the network executive associated with this network model
		virtual netExec* getNetExec() = 0;
		virtual ~devsn()
		{
		}
		devsn* typeIsNetwork() { return this; }
};

/**
Function object that computes the depth of a model in a model hierarchy.
Depth 0 is the top level model, its children have depth 1, etc.
*/
struct compute_model_hier_depth
{
	bool operator()(const devs* A, const devs* B) const
	{
		return depth(A,0) < depth(B,0);
	}
	static int depth(const devs* model, int d)
	{
		if (model->getParent() == NULL) return d;
		return depth(model->getParent(),d+1);
	}
};

ADEVS_NAMESPACE_END

#endif
