/***************
adevs, a library supporting discrete event simulation
Copyright (C) 2003-2004 James Nutaro

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bugs, comments, and questions can be sent to nutaro@ece.arizona.edu
***************/
#include "pdevs.h"
using namespace std;
USING_NAMESPACE_ADEVS

ADEVS_NAMESPACE::pdevs_coord::pdevs_coord(devsn* model, ADEVS_TIME_TYPE t_granule):
basic_sim(t_granule),
top_model(model)
{
	// Schedule all of the component models.  Computes global_tL during scheduling.
	global_tL = ADEVS_ZERO;
	// Models are scheduled recursively using the schedule() method
	schedule(top_model);
	// Set the time of next event
	global_tN = ADEVS_INFINITY;
	if (sched.getSize() > 0)
	{
		global_tN = sched.minPriority();
	}
}

void ADEVS_NAMESPACE::pdevs_coord::schedule(devsn* model)
{
	// Schedule the network executive
	if (model->getNetExec()->timeNext() < ADEVS_INFINITY)
	{
		sched.insert(model->getNetExec(),model->getNetExec()->timeNext());
	}
	if (model->getNetExec()->timeLast() > global_tL) 
	{
		global_tL = model->getNetExec()->timeLast();
	}
	// Schedule each component model
	adevs_set<devs*> components;
	model->getNetExec()->getComponents(components);
	for (int i = 0; i < components.getSize(); i++)
	{
		// Atomic models are placed into the queue if they are not passive
		atomic* amodel = components[i]->typeIsAtomic();
		if (amodel != NULL)
		{
			if (amodel->timeNext() < ADEVS_INFINITY)
			{
				sched.insert(amodel,amodel->timeNext());
			}
			if (amodel->timeLast() > global_tL)
			{
				global_tL = amodel->timeLast();
			}
		}
		// Network models are scheduled recursively
		else
		{
			schedule(components[i]->typeIsNetwork());
		}
	}
}

void ADEVS_NAMESPACE::pdevs_coord::computeInputOutput()
{
	// No imminent models, so no output
	if (sched.getSize() == 0) return;
	// Get all of the imminent models
	while (sched.minPriority() - global_tN <= t_granule)
	{
		imm.add(sched.getMinimum());
		sched.removeMinimum();
		if (sched.getSize() == 0)
		{
			break;
		}
	}
	// Remember the number of real imminents.  We use the fact that elements are
	// always added to the back of the 'set' class to distinguish be actually imminent
	// models and models that are just undergoing an external transition.
	int num_imm = imm.getSize();
	// Compute I/O functions and route events for imminent models.  Save bags of output
	// values for garbage collection at the end of the cycle.
	for (int i = 0; i < num_imm; i++)
	{
		adevs_bag<PortValue>* output_values = io_pool.make_obj();
		// Make sure the output function sees the proper current time
		// (this is for backwards compatibility).
		ADEVS_TIME_TYPE t = imm[i]->timeCurrent();
		imm[i]->setTimeCurrent(imm[i]->timeNext());
		imm[i]->output_func(*output_values);
		// Restore the model's clock (no real advance in time takes place until
		// state transition function is computed).
		imm[i]->setTimeCurrent(t);
		// Send model outputs to their proper destination via the recursive routing
		// function.
		for (int j = 0; j < output_values->getSize(); j++)
		{
			route(imm[i]->getParent(),imm[i],output_values->get(j));
		}
		// Remember output values for later garbage collection.
		imm[i]->set_output_bag(output_values);
	}
}

void ADEVS_NAMESPACE::pdevs_coord::route(devsn* parent, devs* src, PortValue& pv)
{
	adevs_bag<EventReceiver>* recvs = recv_pool.make_obj();
	// Compute the set of receivers for this port value pair
	parent->getNetExec()->route(pv,src,*recvs);
	// Deliver the event to each of its targets
	for (int i = 0; i < recvs->getSize(); i++)
	{
		// Check for self-influencing error condition
		if (src == recvs->get(i).model)
		{
			THROW_BASE_EXCEPTION("Model tried to influence self");
		}
		PortValue pvx(recvs->get(i).port,pv.value);
		// if destination is an atomic model add the event to the IO bag for that model
		// and add model to end of list of imminents for this cycle
		atomic* amodel = recvs->get(i).model->typeIsAtomic();
		if (amodel != NULL)
		{
			if (amodel->get_input_bag() == NULL)
			{
				amodel->set_input_bag(io_pool.make_obj());
				imm.add(amodel);
			}
			amodel->get_input_bag()->add(pvx);
		}
		// otherwise if this is an external output of the top level model, add it
		// to the output table
		else if (recvs->get(i).model == top_model)
		{
			top_model_output.add(pvx);
		}
		// else if this is an external output from the parent model, route it
		// recursively
		else if (recvs->get(i).model == parent)
		{
			route(parent->getParent(),parent,pvx);
		}
		// otherwise, it is an input to a coupled model and should be routed recursively
		else
		{
			route(recvs->get(i).model->typeIsNetwork(),recvs->get(i).model,pvx);
		}
	}
	recvs->removeAll();
	recv_pool.destroy_obj(recvs);
}

void ADEVS_NAMESPACE::pdevs_coord::reschedule_atomic(atomic* model)
{
	// Clean up the input bag for the model
	if (model->get_input_bag() != NULL)
	{
		model->get_input_bag()->removeAll();
		io_pool.destroy_obj(model->get_input_bag());
		model->set_input_bag(NULL);
	}
	// Clean up the output bag for the model
	if (model->get_output_bag() != NULL)
	{
		model->gc_output(*(model->get_output_bag()));
		model->get_output_bag()->removeAll();
		io_pool.destroy_obj(model->get_output_bag());
		model->set_output_bag(NULL);
	}
	// Reschedule the model
	if (model->timeNext() < ADEVS_INFINITY)
	{
		sched.insert(model,model->timeNext());
	}
	else
	{
		sched.remove(model);
	}
}

void ADEVS_NAMESPACE::pdevs_coord::compute_atomic_delta_func(atomic* model, 
ADEVS_TIME_TYPE t)
{
	adevs_bag<PortValue>* x = model->get_input_bag();
	model->setTimeCurrent(t);
	if (model->sigma() <= t_granule && x == NULL)
	{
		model->delta_int();
		model->setTimeLast(t);
	}
	else if (model->sigma() <= t_granule && x != NULL)
	{
		model->delta_conf(*x);
		model->setTimeLast(t);
	}
	else if (model->sigma() > t_granule && x != NULL)
	{
		model->delta_ext(model->elapsed(),*x);
		model->setTimeLast(t);
	}
	if (x != NULL)
	{
		x->removeAll();
		io_pool.destroy_obj(x);
		model->set_input_bag(NULL);
	}
}

void ADEVS_NAMESPACE::pdevs_coord::computeDeltaFunc(ADEVS_TIME_TYPE t)
{
	// index variables
	int i;
	// Check the requested time of next event for validity
	if (t < global_tL || t > global_tN + t_granule)
	{
		THROW_BASE_EXCEPTION("Invalid time of next event");
	}
	// If t is too early, we have no imminent models
	// (remember, computeInputOutput assumes t = tN)
	// so place all models imminent at tN back into the
	// schedule
	if (global_tN - t > t_granule)
	{
		for (i = 0; i < imm.getSize(); i++)
		{
			reschedule_atomic(imm[i]);
		}
		imm.removeAll();
	}
	// Route external input
	for (i = 0; i < top_model_input.getSize(); i++)
	{
		route(top_model,top_model,top_model_input[i]);
	}
	/**
	Compute next state for all imminent models.  If the model is a network executive,
	construct the new model and deleted model sets as well.  New models are added
	to the end of the immininent set and will be scheduled at the end of the simulation
	cycle.
	*/
	int net_exec_count = 0;
	for (i = 0; i < imm.getSize(); i++)
	{
		if (imm[i]->typeIsNetExec() != NULL)
		{
			net_exec_count++;
		}
		else
		{
			compute_atomic_delta_func(imm[i],t);
		}
	}
	/// Compute new states for network executive models and apply structure changes
	compute_structure_changes(net_exec_count,t);
	/// Set time of last event
	global_tL = t;
}

/**
Recursively removes a model from the simulator.
*/
void ADEVS_NAMESPACE::pdevs_coord::remove_model(devs* model)
{
	// If the model is atomic, remove it from the schedule, recover any
	// I/O bags being used by the model, and remove it from the imminent set.
	if (model->typeIsAtomic())
	{
		model->typeIsAtomic()->setTimeNext(ADEVS_INFINITY);
		reschedule_atomic(model->typeIsAtomic());
		if (imm.contains(model->typeIsAtomic()))
		{
			imm.erase(model->typeIsAtomic());
		}
	}
	// Otherwise, get the components and remove them recursively
	else
	{
		adevs_set<devs*> components;
		model->typeIsNetwork()->getNetExec()->getComponents(components);
		for (int i = 0; i < components.getSize(); i++)
		{
			remove_model(components[i]);
		}
		remove_model(model->typeIsNetwork()->getNetExec());
	}
}

/**
Computes structure changes.  New models are initialized and scheduled.
Old models are removed from simulator and deleted.
*/
void ADEVS_NAMESPACE::pdevs_coord::compute_structure_changes(int net_exec_count, 
ADEVS_TIME_TYPE t)
{
	if (net_exec_count > 0)
	{
		// Sort imminent network executives by depth in hierarchy
		vector<netExec*> nx(net_exec_count);
		for (int i = 0; net_exec_count > 0; i++)
		{
			nx[net_exec_count-1] = imm[i]->typeIsNetExec();
			if (nx[net_exec_count-1] != NULL)
			{
				net_exec_count--;
			}
		}
		sort(nx.begin(),nx.end(),compute_model_hier_depth());
		// Compute state and structure changes in depth order
		adevs_set<devs*> s1, s2, s3;
		for (vector<netExec*>::iterator j = nx.begin(); j != nx.end(); j++)
		{
			// Make sure model still exists!
			if (imm.contains(*j))
			{
				// Get the structure change hint
				bool structure_change = (*j)->componentSetChanged();
				// Do structure change unless hint says not to do it
				if (structure_change)
				{
					// Get component set before the state transition
					(*j)->getComponents(s1);
				}
					// Compute the next state
					compute_atomic_delta_func(*j,t);
				// Finish structure change according to the hint prior to the
				// state transition
				if (structure_change)
				{
					// Get the new component set
					(*j)->getComponents(s2);
					// Indexing variable
					int k;
					// Destroy all old models
					s3 = s1-s2;
					for (k = 0; k < s3.getSize(); k++)
					{
						remove_model(s3[k]);
					}
					(*j)->gc_models(s3);	
					// Initialize and schedule all of the new models
					s3 = s2-s1;
					for (k = 0; k < s3.getSize(); k++)
					{
						s3[k]->setParent((*j)->getParent());
						s3[k]->init(t);
						atomic* amodel = s3[k]->typeIsAtomic();
						if (amodel != NULL)
						{
							if (amodel->timeNext() < ADEVS_INFINITY)
							{
								sched.insert(amodel,amodel->timeNext());
							}
						}
						else
						{
							schedule(s3[k]->typeIsNetwork());
						}
					}
				}
			}
		}
	}
}

void ADEVS_NAMESPACE::pdevs_coord::endOfCycle()
{
	// Cleanup IO and reschedule all models that changed stated
	for (int i = 0; i < imm.getSize(); i++)
	{
		// Give models a chance to clean up their output values for this cycle
		if (imm[i]->get_output_bag() != NULL)
		{
			imm[i]->gc_output(*(imm[i]->get_output_bag()));
			imm[i]->get_output_bag()->removeAll();
			io_pool.destroy_obj(imm[i]->get_output_bag());
			imm[i]->set_output_bag(NULL);
		}
		// Schedule the models that changed state this cycle
		if (imm[i]->timeNext() < ADEVS_INFINITY)
		{
			sched.insert(imm[i],imm[i]->timeNext());
		}
		else 
		{
			sched.remove(imm[i]);
		}
	}
	imm.removeAll();
	top_model_input.removeAll();
	top_model_output.removeAll();
	global_tN = ADEVS_INFINITY;
	if (sched.getSize() > 0)
	{
		global_tN = sched.minPriority();
	}
}

ADEVS_NAMESPACE::pdevs_coord::~pdevs_coord()
{
}
